#!/usr/bin/python3

from collections.abc import Iterator, Sequence
import itertools
from typing import Any, TypeVar
#>from typing import Self # from v3.11!

from .abc import Array
from .array1d import Array1D
from .format import *

T = TypeVar('T')
Boundaries = tuple[tuple[int, int], ...]

def _minmax(arr: tuple[tuple[int, int], ...]) -> tuple[int, int]:
    minarr, maxarr = zip(*arr)
    return min(minarr), max(maxarr)


class ArrayND(Array):
    def __init__(self,
            dim: int,
            default: T|None = None,
            *,
            content: Sequence|None = None,
            offset: tuple[int,...]|list[int]|int = 0
            ) -> None:
        self._pos: list = [] # list[Self|Array1D]
        self._neg: list = [] # list[Self|Array1D]
        self._dim: int = dim
        self._default: Any = default
        if content is not None:
            self.replace_content(content, offset)
        self.index_format = None


    @property
    def dim(self) -> int:
        return self._dim

    @property
    def offset(self) -> tuple[int, ...]:
        return tuple(map(lambda e: e[0], self.boundaries))

    @property
    def shape(self) -> tuple[int, ...]:
        return tuple(map(lambda e: e[1] - e[0], self.boundaries))

    @property
    def boundaries(self) -> Boundaries:
        if len(self) == 0:
            return (((0, 0),) * self._dim)
        all_bounds: Iterator[Boundaries]
        if self._dim == 2:
            all_bounds = ((plane.boundaries,) for plane in self)
        else:
            all_bounds = (plane.boundaries for plane in self)
        boundmax: Iterator[tuple[int, int]] = (_minmax(a) for a in zip(*all_bounds))
        return ((-len(self._neg), len(self._pos)), *boundmax)


    def replace_content(self, array: Sequence,
                        offset: tuple[int,...]|list[int]|int = 0) -> None:
        self._neg = []
        self._pos = []
        if isinstance(offset, int):
            offset = [offset] * self._dim
        offset = list(offset)
        current_offset: int = 0
        sub_offset: list[int] = [0]
        if len(offset) > 0:
            current_offset = offset[0]
            if len(offset) > 1:
                sub_offset = offset[1:]
        for index, subarray in enumerate(array, current_offset):
            plane = self._getplane(index) # Self|Array1D
            if self._dim == 2:
                plane.replace_content(subarray, sub_offset[0])
            else:
                plane.replace_content(subarray, sub_offset)

    def __setitem__(self, index: tuple[int, ...], value: T) -> None:
        if not isinstance(index, tuple) or len(index) != self._dim \
                or any(map(lambda x: not isinstance(x, int), index)):
            raise TypeError('Index must be a {self._dim} element tuple of integers')
        plane = self._getplane(index[0]) # Self|Array1D
        if self._dim == 2:
            plane[index[1]] = value
        else:
            plane[index[1:]] = value

    def __getitem__(self, index: int|tuple[int, ...]|slice) -> Any: # Self|Array1D|T
        if isinstance(index, slice):
            range_indices = self._range_indices(index)
            # Return iterator instead of some arbitrary collection
            return (self._getplane(i) for i in range(*range_indices))
        if isinstance(index, int):
            return self._getplane(index)
        if not isinstance(index, tuple) or len(index) != self._dim \
                or any(map(lambda x: not isinstance(x, int), index)):
            raise TypeError('Index must be a {self._dim} element tuple of integers')
        plane = self._getplane(index[0], create=False)
        if plane is None:
            return self._default
        elif self._dim == 2:
            return plane[index[1]]
        else:
            return plane[index[1:]]

    def __iter__(self) -> itertools.chain:
        return itertools.chain(reversed(self._neg), self._pos)

    def __len__(self) -> int:
        return len(self._pos) + len(self._neg)

    def __format__(self, format: str) -> str:
        formatter: Formatter = Formatter(self._default)
        if self.index_format:
            formatter.index_format = self.index_format
        formatter.apply_format_string(format)
        return self._format(formatter)

    def __str__(self) -> str:
        return self._format(StrFormatter(self._default))

    def __repr__(self) -> str:
        repr_string: str = self._format(ReprFormatter(self._default))
        if repr_string != '[]':
            repr_string = '\n' + repr_string
        return f'ArrayND(dim={self._dim}, default={self._default!r}, ' \
            f'offset={self.offset}, content={repr_string})'


    def _range_indices(self, indices: slice) -> tuple[int, int, int]:
        range_indices: list[int|None] = [indices.start, indices.stop, indices.step]
        if range_indices[2] is None:
            range_indices[2] = 1
        assert isinstance(range_indices[2], int)
        if range_indices[0] is None:
            if range_indices[2] > 0:
                range_indices[0] = -len(self._neg)
            else:
                range_indices[0] = len(self._pos) - 1
        if range_indices[1] is None:
            if range_indices[2] > 0:
                range_indices[1] = len(self._pos)
            else:
                range_indices[1] = -len(self._neg) - 1
        assert isinstance(range_indices[0], int)
        assert isinstance(range_indices[1], int)
        return (range_indices[0], range_indices[1], range_indices[2])

    def _getplane(self, index: int, create: bool = True) -> Any: # Self|Array1D
        if index >= 0:
            part = self._pos
        else:
            part = self._neg
            index = -index - 1
        if len(part) <= index:
            if not create:
                return None
            if self._dim == 2:
                part.extend([
                    Array1D(default=self._default)
                        for _ in range(index - len(part) + 1)
                ])
            else:
                part.extend([
                    ArrayND(dim=self._dim - 1, default=self._default)
                        for _ in range(index - len(part) + 1)
                ])
        return part[index]

    def _maxwidth(self, formatter: Formatter, boundaries: Boundaries) -> None:
        # This private method assumes, that repr shows all values
        for plane in self:
            plane._maxwidth(formatter, boundaries[1:])
        if boundaries[0][0] < -len(self._neg) \
                or boundaries[0][1] > len(self._pos):
            formatter.update_maxwidth_default()

    def _output(self, formatter: Formatter, boundaries: Boundaries,
                        indent: str = '', indices: list[int] = []) -> None:
        continued: bool = False
        separator: str = '\n' * (self._dim-2)
        subindent: str = indent + ' '
        dummy: Array1D|ArrayND|None = None
        formatter.output_begin()
        for index in range(boundaries[0][0], boundaries[0][1]):
            if continued:
                if self._dim == 3:
                    formatter.output_rowsep(separator, subindent, indices + [index])
                else:
                    formatter.output_rowsep(separator, subindent)
            else:
                if self._dim == 3:
                    formatter.output_firstrow(subindent, indices + [index])
            continued = True
            plane = self._getplane(index, create = False)
            if plane is None:
                if dummy is None: # lazy evaluation if needed
                    if self._dim == 2:
                        dummy = Array1D(self._default)
                    else:
                        dummy = ArrayND(self._dim - 1, self._default)
                plane = dummy
            plane._output(formatter, boundaries[1:], subindent, indices + [index])
        formatter.output_end()

    def _format(self, formatter: Formatter) -> str:
        boundaries = self.boundaries
        self._maxwidth(formatter, boundaries)
        self._output(formatter, boundaries)
        return formatter.output



import pytest

from stretchy import Array1D

@pytest.mark.parametrize('default',
    (42, '#', 42.69, None, False)
)
def test_default(default):
    s = Array1D(default)
    assert s[2] == default
    assert s[-2] == default
    s[5] = ...
    assert s[3] == default
    s[-5] = ...
    assert s[-3] == default


def test_replace_content():
    # Simple prelimunary test; more tests later...
    s = Array1D('.')
    s.replace_content('abcde', 2)
    assert f'{s:s}' == '..abcde'
    s.replace_content('abcde', -7)
    assert f'{s:s}' == 'abcde..'


@pytest.mark.parametrize('pos, content',
    (
        ((0,), '#'),
        ((2,), '..#'),
        ((6,2), '..#...#'),
        ((2,6), '..#...#'),
        ((-2,), '#.'),
        ((-6,-2), '#...#.'),
        ((-2,-6), '#...#.'),
        ((-2,2), '#...#'),
        ((2,-2), '#...#'),
    )
)
def test_setitem(pos, content):
    s = Array1D('.')
    for p in pos:
        s[p] = '#'
    assert f'{s:s}' == content


@pytest.mark.parametrize('default, arr, check',
    (
        ([], ((None,None,7),), {0:None, 1:None, 2:7, 3:None, -1:None}),
        ((0,), ((0,0,7),), {0:0, 1:0, 2:7, 3:0, -1:0}),
        ((0,), ((7,),), {0:7, 1:0, -1:0}),
        ((0,), ((7,),2), {0:0, 1:0, 2:7, 3:0, -1:0}),
        ((0,), ((7,),-2), {0:0, -1:0, -2:7, -3:0, 1:0}),
    )
)
def test_getitem(default, arr, check):
    s = Array1D(*default)
    s.replace_content(*arr)
    for pos, value in check.items():
        assert s[pos] == value


SLICE_INPUT = (
    ((None,), '#########', 'abcdefghi'),

    ((-2,None), 'ab#######', 'cdefghi'),
    ((0,None), 'abcd#####', 'efghi'),
    ((2,None), 'abcdef###', 'ghi'),
    ((-6,None), '###########', '..abcdefghi'),
    ((6,None), 'abcdefghi', ''),

    ((None,-2), '##cdefghi', 'ab'),
    ((None,0), '####efghi', 'abcd'),
    ((None,2), '######ghi', 'abcdef'),
    ((None,-6), 'abcdefghi', ''),
    ((None,7), '###########', 'abcdefghi..'),

    ((-2,2), 'ab####ghi', 'cdef'),
    ((-6,2), '########ghi', '..abcdef'),
    ((-2,7), 'ab#########', 'cdefghi..'),
    ((-6,7), '#############', '..abcdefghi..'),

    ((None,None,3), '#bc#ef#hi', 'adg'),
    ((None,None,-3), 'ab#de#gh#', 'ifc'),
    ((-6,7,3), '#.a#cd#fg#i.#', '.beh.'),
    ((6,-7,-3), '#.a#cd#fg#i.#', '.heb.'),
)

@pytest.mark.parametrize('indices, content, got', SLICE_INPUT)
def test_setitem_slice(indices, content, got):
    s = Array1D(default='.', content='abcdefghi', offset=-4)
    s[slice(*indices)] = '#'
    assert f'{s:s}' == content


@pytest.mark.parametrize('indices, content, got', SLICE_INPUT)
def test_getitem_slice(indices, content, got):
    s = Array1D(default='.', content='abcdefghi', offset=-4)
    assert ''.join(s[slice(*indices)]) == got


# Followings test also the `replace_content` method
TEST_DATA = (
    (((7,),), [7]),
    (((7,),2), [None, None, 7]),
    (((7,),-2), [7, None]),
    (((7,8,9),2), [None, None, 7, 8, 9]),
    (((7,8,9),-2), [7, 8, 9]),
)

@pytest.mark.parametrize('arr, content', TEST_DATA)
def test_offset(arr, content):
    s = Array1D()
    s.replace_content(*arr)
    offset = arr[1] if len(arr) == 2 and arr[1] < 0 else 0
    assert s.offset == offset


@pytest.mark.parametrize('arr, content', TEST_DATA)
def test_boundaries(arr, content):
    s = Array1D()
    s.replace_content(*arr)
    offset = arr[1] if len(arr) == 2 and arr[1] < 0 else 0
    assert s.boundaries == (offset, offset + len(content))


@pytest.mark.parametrize('arr, content', TEST_DATA)
def test_len(arr, content):
    s = Array1D()
    s.replace_content(*arr)
    assert len(s) == len(content)


@pytest.mark.parametrize('arr, content', TEST_DATA)
def test_iter(arr, content):
    s = Array1D()
    s.replace_content(*arr)
    assert list(s) == content


@pytest.mark.parametrize('params, offset, content',
    (
        (('12345',), 0, '12345'),
        (('12345',3), 0, '...12345'),
        (('12345',-3), -3, '12345'),
        (('12345',-7), -7, '12345..'),
        (((1,2,3,4,5),-3), -3, '12345'),
        (([1,2,3,4,5],-3), -3, '12345'),
    )
)
def test_set(params, offset, content):
    s = Array1D('.')
    s.replace_content(*params)
    assert s.offset == offset
    assert f'{s:s}' == content


@pytest.fixture
def array():
    s = Array1D(default='.', offset=-3,
        content=('x', None,'.',234,'.',False,6.7))
    return s


def test_str_empty():
    s = Array1D()
    assert str(s) == "[]"


def test_str(array):
    assert str(array) == "[x           .       234 .     False   6.7]"


def test_repr_empty():
    s = Array1D()
    assert repr(s) == "Array1D(default=None, offset=0, content=[])"


def test_repr(array):
    assert repr(array) == "Array1D(default='.', offset=-3, " \
        "content=['x'  , None , '.'  ,   234, '.'  , False,   6.7])"


@pytest.mark.parametrize('fmt,result',
    (
        ('{}', "x  . 234 . False 6.7"),
        ('{:}', "x  . 234 . False 6.7"),
        ('{!s}', "[x           .       234 .     False   6.7]"),
        ('{!r}', "Array1D(default='.', offset=-3, " \
            "content=['x'  , None , '.'  ,   234, '.'  , False,   6.7])"),

        ('{:a}', "x           .       234 .     False   6.7"),

        ('{:s}', "x.234.False6.7"),
        ('{:s }', "x  . 234 . False 6.7"),
        ('{:s,}', "x,,.,234,.,False,6.7"),

        ('{:l}', "'x' None '.' 234 '.' False 6.7"),
        ('{:ls}', "'x'None'.'234'.'False6.7"),
        ('{:ls }', "'x' None '.' 234 '.' False 6.7"),
        ('{:ls,}', "'x',None,'.',234,'.',False,6.7"),

        ('{:s,b<e>}', "<x,,.,234,.,False,6.7>"),
    )
)
def test_format(fmt, result, array):
    assert fmt.format(array) == result


def test_wrong_format(array):
    with pytest.raises(ValueError):
        '{:@@@}'.format(array)




#!/usr/bin/python3

from collections.abc import Iterable
from functools import partial
from io import StringIO
from typing import Any


def _valwidth_str(item: Any) -> int:
    return len(str(item)) if item is not None else 0

def _valwidth_repr(item: Any) -> int:
    return len(repr(item))


def _valrepr_str(item: Any, width: int) -> str:
    if item is None:
        return ' ' * width
    if isinstance(item, (int, float)) and not isinstance(item, bool):
        return f'{item!s: >{width}}'
    else:
        return f'{item!s: <{width}}'

def _valrepr_repr(item: Any, width: int) -> str:
    if isinstance(item, (int, float)) and not isinstance(item, bool):
        return f'{item!r: >{width}}'
    else:
        return f'{item!r: <{width}}'


class Formatter:
    sep: str = ' '
    rowend: str = ''
    begin: str = ''
    end: str = ''
    index: bool = False
    arrange: bool = False
    _literal: bool = False
    index_format: str = 'Index {}:'

    def __init__(self, default: Any = None) -> None:
        self._default: Any = default
        self.reset()
        self.literal = self._literal

    @property
    def literal(self) -> bool:
        return self._literal

    @literal.setter
    def literal(self, value: bool) -> None:
        self._literal = value
        if value:
            self._valwidth = _valwidth_repr
            self._valrepr = _valrepr_repr
        else:
            self._valwidth = _valwidth_str
            self._valrepr = _valrepr_str

    @property
    def output(self) -> str:
        return self._output.getvalue()

    def reset(self) -> None:
        self._maxwidth: int = 0
        self._output: StringIO = StringIO()

    def update_maxwidth(self, content: Iterable) -> None:
        maxwidth: int = max(map(self._valwidth, content))
        self._maxwidth = max(self._maxwidth, maxwidth)

    def update_maxwidth_default(self) -> None:
        self._maxwidth = max(self._maxwidth, self._valwidth(self._default))

    def output_iter(self, content: Iterable) -> None:
        self._output.write(self.begin)
        if self.arrange:
            maxwidth = self._maxwidth
        else:
            maxwidth = 0
        valrepr = partial(self._valrepr, width=maxwidth)
        itrepr: str = self.sep.join(map(valrepr, content))
        self._output.write(itrepr)
        self._output.write(self.end)

    def output_begin(self) -> None:
        self._output.write(self.begin)

    def output_end(self) -> None:
        self._output.write(self.end)

    def output_firstrow(self, indent: str, index: list[int] = []) -> None:
        if not self.index or not index:
            return
        if self.begin:
            self._output.write('\n')
        self._output.write(self.index_format.format(','.join(map(str,index))))
        self._output.write('\n')
        if self.begin:
            self._output.write(indent)

    def output_rowsep(self, separator: str, indent: str, index: list[int] = []) -> None:
        self._output.write(self.rowend + '\n')
        if self.index:
            if index:
                self._output.write(self.index_format.format(','.join(map(str,index))))
                self._output.write('\n')
        else:
            self._output.write(separator)
        if self.begin:
            self._output.write(indent)

    def output_string(self, content: str) -> None:
        self._output.write(content)

    def apply_format_string(self, format_string: str) -> None:
        current = None
        for c in format_string:
            if c == 's':
                self.sep = ''
                current = 'sep'
            elif c == 'r':
                self.rowend = ''
                current = 'rowend'
            elif c == 'b':
                self.begin = ''
                current = 'begin'
            elif c == 'e':
                self.end = ''
                current = 'end'
            elif c == 'i':
                self.index = True
                current = None
            elif c == 'a':
                self.arrange = True
                current = None
            elif c == 'l':
                self.literal = True
                current = None
            elif current == 'sep':
                self.sep += c
            elif current == 'rowend':
                self.rowend += c
            elif current == 'begin':
                self.begin += c
            elif current == 'end':
                self.end += c
            else:
                raise ValueError(f"Unknown format code '{c}' stretchy object")


class StrFormatter (Formatter):
    begin = '['
    end = ']'
    arrange = True


class ReprFormatter (Formatter):
    sep = ', '
    rowend = ','
    begin = '['
    end = ']'
    arrange = True
    _literal = True



#!/usr/bin/python3

import pytest
from itertools import product

import stretchy


@pytest.mark.parametrize('array, atype',
    (
        (stretchy.empty(1), stretchy.Array1D),
        (stretchy.empty(2), stretchy.ArrayND),
        (stretchy.empty(3), stretchy.ArrayND),
        (stretchy.array('abcde'), stretchy.Array1D),
        (stretchy.array(['ab','de']), stretchy.Array1D),
        (stretchy.array(['ab','de'], dim=2), stretchy.ArrayND),
        (stretchy.array([['ab','cd'],['ef','gh']]), stretchy.ArrayND),
    )
)
def test_isinstance(array, atype):
    assert isinstance(array, atype)
    assert isinstance(array, stretchy.Array)
    assert type(array) is atype


@pytest.mark.parametrize('params, dim, default, content, offset',
    (
        # dim
        ({}, 1, None, '', 0),
        ({'dim':1}, 1, None, '', 0),
        ({'dim':2}, 2, None, '', (0,)*2),
        ({'dim':4}, 4, None, '', (0,)*4),
        # default
        ({'default':'x'}, 1, 'x', '', 0),
        ({'dim':2, 'default':'x'}, 2, 'x', '', (0,)*2),
        ({'dim':4, 'default':'x'}, 4, 'x', '', (0,)*4),
        # content
        ({'content':'abcde'}, 1, None, 'a,b,c,d,e', 0),
        ({'content':['ab','cd']}, 1, None, 'ab,cd', 0),
        ({'content':['ab','cd'], 'dim':2}, 2, None, 'a,b\nc,d', (0,)*2),
        ({'content':[['ab','cd'],['ef','gh']]}, 2, None, 'ab,cd\nef,gh', (0,)*2),
        ({'content':[['ab','cd'],['ef','gh']], 'dim':3}, 3, None, 'a,b\nc,d\n\ne,f\ng,h', (0,)*3),
        ({'content':range(-4,5,2)}, 1, None, '-4,-2,0,2,4', 0),
        ({'content':map(''.join, product('xy',repeat=2))}, 1, None, 'xx,xy,yx,yy', 0),
        # offset
        ({'content':'abcde', 'offset':-2}, 1, None, 'a,b,c,d,e', -2),
        ({'content':'abcde', 'offset':-8}, 1, None, 'a,b,c,d,e,,,', -8),
        ({'content':'abcde', 'offset':0}, 1, None, 'a,b,c,d,e', 0),
        ({'content':'abcde', 'offset':3}, 1, None, ',,,a,b,c,d,e', 0),
        ({'content':['ab','cd'], 'dim':2, 'offset':-1}, 2, None, 'a,b\nc,d', (-1,)*2),
        ({'content':['ab','cd'], 'dim':2, 'offset':-3}, 2, None, 'a,b,\nc,d,\n,,', (-3,)*2),
        ({'content':['ab','cd'], 'dim':2, 'offset':0}, 2, None, 'a,b\nc,d', (0,)*2),
        ({'content':['ab','cd'], 'dim':2, 'offset':1}, 2, None, ',,\n,a,b\n,c,d', (0,)*2),
        ({'content':['ab','cd'], 'dim':2, 'offset':(-3,-4)}, 2, None, 'a,b,,\nc,d,,\n,,,', (-3,-4)),
        ({'content':['ab','cd'], 'dim':2, 'offset':(-3,2)}, 2, None, ',,a,b\n,,c,d\n,,,', (-3,0)),
        ({'content':['ab','cd'], 'dim':2, 'offset':(1,-4)}, 2, None, ',,,\na,b,,\nc,d,,', (0,-4)),
        ({'content':['ab','cd'], 'dim':2, 'offset':(1,2)}, 2, None, ',,,\n,,a,b\n,,c,d', (0,)*2),
    )
)
def test_array(params, dim, default, content, offset):
    array = stretchy.array(**params)
    assert array.dim == dim
    assert array._default == default
    if dim == 1:
        pos = 200
    else:
        pos = (200,) * dim
    assert array[pos] == default
    assert array.offset == offset
    assert f'{array:s,}' == content


@pytest.mark.parametrize('params, dim, default',
    (
        (tuple(), 1, None),
        ((1,), 1, None),
        ((2,), 2, None),
        ((4,), 4, None),
        ((1,'x'), 1, 'x'),
        ((2,'x'), 2, 'x'),
        ((4,'x'), 4, 'x'),
    )
)
def test_empty(params, dim, default):
    array = stretchy.empty(*params)
    assert array.dim == dim
    assert array._default == default
    if dim == 1:
        pos = 2
    else:
        pos = (2,) * dim
    assert array[pos] == default



#!/usr/bin/python3

import itertools
from typing import Any, Callable, TypeVar
from collections.abc import Iterable, Iterator

from .abc import Array
from .format import *

T = TypeVar('T')
Boundaries = tuple[tuple[int, int], ...]

class Array1D(Array):
    def __init__(self,
            default: T|None = None,
            *,
            content: Iterable|None = None,
            offset: int = 0
            ) -> None:
        self._pos: list[T|None] = []
        self._neg: list[T|None] = []
        self._default: T|None = default
        if content is not None:
            self.replace_content(content, offset)


    @property
    def dim(self) -> int:
        return 1

    @property
    def offset(self) -> int:
        return -len(self._neg)

    @property
    def boundaries(self) -> tuple[int, int]:
        return -len(self._neg), len(self._pos)


    def replace_content(self, content: Iterable, offset: int = 0) -> None:
        if offset >= 0:
            self._neg = []
            self._pos = [self._default] * offset + list(content)
            return
        self._neg = [self._default] * -offset
        self._pos = []
        it = iter(content)
        try:
            for index in range(-offset-1,-1,-1):
                self._neg[index] = next(it)
            while True:
                self._pos.append(next(it))
        except StopIteration:
            return


    def __setitem__(self, index: int|slice, value: T|None) -> None:
        dim: list[int|None] = [None, None]
        if isinstance(index, slice):
            range_indices = self._range_indices(index)
            # Fill with value! (Python collections do not support this)
            for i in range(*range_indices):
                self.__setitem__(i, value)
            return
        if index >= 0:
            if len(self._pos) <= index:
                self._pos.extend([self._default] * (index - len(self._pos) + 1))
                dim[1] = len(self._pos)
            self._pos[index] = value
        else:
            index = -index - 1
            if len(self._neg) <= index:
                self._neg.extend([self._default] * (index - len(self._neg) + 1))
                dim[0] = -len(self._neg)
            self._neg[index] = value

    def __getitem__(self, index: int|slice) -> T|Iterator|None:
        if isinstance(index, slice):
            range_indices = self._range_indices(index)
            # Return iterator instead of some arbitrary collection
            return (self.__getitem__(i) for i in range(*range_indices))
        if index >= 0:
            if len(self._pos) <= index:
                return self._default
            return self._pos[index]
        else:
            index = -index - 1
            if len(self._neg) <= index:
                return self._default
            return self._neg[index]

    def __iter__(self) -> itertools.chain:
        return itertools.chain(reversed(self._neg), self._pos)

    def __len__(self) -> int:
        return len(self._pos) + len(self._neg)

    def __format__(self, format: str) -> str:
        formatter: Formatter = Formatter(self._default)
        formatter.apply_format_string(format)
        return self._format(formatter)

    def __str__(self) -> str:
        return self._format(StrFormatter(self._default))

    def __repr__(self) -> str:
        repr_string: str = self._format(ReprFormatter(self._default))
        return f'Array1D(default={self._default!r}, ' \
            f'offset={self.offset}, content={repr_string})'


    def _range_indices(self, indices: slice) -> tuple[int, int, int]:
        range_indices: list[int|None] = [indices.start, indices.stop, indices.step]
        if range_indices[2] is None:
            range_indices[2] = 1
        assert isinstance(range_indices[2], int)
        if range_indices[0] is None:
            if range_indices[2] > 0:
                range_indices[0] = -len(self._neg)
            else:
                range_indices[0] = len(self._pos) - 1
        if range_indices[1] is None:
            if range_indices[2] > 0:
                range_indices[1] = len(self._pos)
            else:
                range_indices[1] = -len(self._neg) - 1
        assert isinstance(range_indices[0], int)
        assert isinstance(range_indices[1], int)
        return (range_indices[0], range_indices[1], range_indices[2])

    def _maxwidth(self, formatter: Formatter, boundaries: Boundaries) -> None:
        # This private method assumes, that repr shows all values
        if len(self):
            formatter.update_maxwidth(self)
        if boundaries[0][0] < -len(self._neg) \
                or boundaries[0][1] > len(self._pos):
            formatter.update_maxwidth_default()

    def _output(self, formatter: Formatter, boundaries: Boundaries, indent: str = '', indices: list[int] = []) -> None:
        b: tuple[int, int] = self.boundaries
        pre: int = b[0] - boundaries[0][0]
        post: int = boundaries[0][1] - b[1]
        items: itertools.chain = itertools.chain(
            itertools.repeat(self._default, pre),
            self,
            itertools.repeat(self._default, post),
        )
        formatter.output_iter(items)

    def _format(self, formatter: Formatter) -> str:
        if len(self):
            formatter.update_maxwidth(self)
        formatter.output_iter(self)
        return formatter.output



#!/usr/bin/python3

from typing import Any
from collections.abc import Iterable, Sequence

from .abc import Array
from .array1d import Array1D
from .arraynd import ArrayND


def _array_dim(content: Sequence, dim: int = 1) -> int:
    dims: list[int] = []
    for item in content:
        if isinstance(item, (list, tuple)):
            dims.append(_array_dim(item, dim + 1))
        else:
            return dim
    if dims:
        return max(dims)
    return dim


def array(
        content: Sequence|Iterable|None = None,
        *,
        default: Any = None,
        offset: tuple[int, ...]|list[int]|int = 0,
        dim: int|None = None
        ) -> Array1D|ArrayND:
    if dim is None:
        if not content or not isinstance(content, Sequence) \
                or isinstance(content, str):
            dim = 1
        else:
            dim = _array_dim(content)
    assert dim > 0
    if dim == 1:
        assert isinstance(offset, int)
        return Array1D(default=default, content=content, offset=offset)
    else:
        assert isinstance(content, (Sequence, type(None)))
        return ArrayND(dim=dim, default=default, content=content, offset=offset)


def empty(dim: int = 1, default: Any = None) -> Array1D|ArrayND:
    if dim == 1:
        return Array1D(default)
    else:
        return ArrayND(dim, default)
